;===============================================================================
;
;  function LOADMHDHDF reads scientific data produced by Kin-MHD in
;  HDF format.
;
;-------------------------------------------------------------------------------
;
;  arguments:
;
;    file - file to read data from
;    var  - variable to be readed
;
function loadmhdhdf, file, var, xcoord = x, ycoord = y, zcoord = z, time = t, single = single, info = info

  data = -1
  ncpus = 1

; check it the file is really written in HDF format
;
  if (not hdf_ishdf(file)) then begin
    print, 'This is not an HDF file or this file does not exist!'
    return, data
  endif

; start the SD interface
;
  sd_id = hdf_sd_start(file, /read)

; check if there were problems with opening the SD interface
;
  if (sd_id eq -1) then begin
    print, 'Could not initiate the SD interface for this file. the file might be corrupted!'

; end SD interface
;
    hdf_sd_end, sd_id

; return -1
;
    return, -1
  endif

; seems that eferything went fine, so proceed with the file
;
; if keyword /LIST is defined list all variables only
;
  if (keyword_set(info)) then begin

; get number of datasets and attributes
;
    hdf_sd_fileinfo, sd_id, ndatasets, nattributes

; print all attribute names
;
    print, 'Number of attributes:', nattributes
    attrs = strarr(nattributes)
    print, 'Attribute names:'
    for n = 0, nattributes - 1 do begin
      hdf_sd_attrinfo, sd_id, n, name = nm
      attrs[n] = nm
    endfor
    print, '    ', attrs
    print, ''

; print all dataset names
;
    print, 'Number of datasets  :', ndatasets
    names = strarr(ndatasets)
    print, 'Dataset names:'
    for n = 0, ndatasets - 1 do begin
      sds_id = hdf_sd_select(sd_id, n)
      hdf_sd_getinfo, sds_id, name = nm
      names[n] = nm
      hdf_sd_endaccess, sds_id
    endfor
    print, '    ', names
    print, ''

; return 0
;
    return, 0
  endif

; read time
;
  attr_index = hdf_sd_attrfind(sd_id, 'time')
  if (attr_index ne -1) then hdf_sd_attrinfo, sd_id, attr_index, data = t
  if (size(t, /dimensions) eq 1) then t = t[0]

; check if this is only a part of data (calculated on multi processor machines)
;
  attr_index = hdf_sd_attrfind(sd_id, 'ncpus')
  if (attr_index ne -1) then hdf_sd_attrinfo, sd_id, attr_index, data = ncpus

; proceed if ncpus == 1 or keword /SINGLE is set
;
  if (ncpus eq 1 or attr_index eq -1 or keyword_set(single)) then begin  ; data in a single file

; find the index of requested dataset
;
    sds_index = hdf_sd_nametoindex(sd_id, var)

    if (sds_index eq -1) then begin

      print, 'Could not find variable ', var, ' in the file.'
      print, 'Please add keyword /INFO to see all stored variables.'

    endif else begin

; select correct dataset
;
      sds_id = hdf_sd_select(sd_id, sds_index)

; read data from this dataset
;
      hdf_sd_getdata, sds_id, data

; and access to this dataset
;
      hdf_sd_endaccess, sds_id

; read coordinates
;
      attr_index = hdf_sd_attrfind(sd_id, 'x')
      if (attr_index ne -1) then hdf_sd_attrinfo, sd_id, attr_index, data = x
      attr_index = hdf_sd_attrfind(sd_id, 'y')
      if (attr_index ne -1) then hdf_sd_attrinfo, sd_id, attr_index, data = y
      attr_index = hdf_sd_attrfind(sd_id, 'z')
      if (attr_index ne -1) then hdf_sd_attrinfo, sd_id, attr_index, data = z

    endelse

; end SD interface
;
    hdf_sd_end, sd_id

  endif else begin                                                    ; data in multiple files

; find the index of requested dataset
;
    sds_index = hdf_sd_nametoindex(sd_id, var)

    if (sds_index eq -1) then begin

      print, 'Could not find variable ', var, ' in the file.'
      print, 'Please add keyword /INFO to see all stored variables.'

; end SD interface
;
      hdf_sd_end, sd_id

; return -1
;
      return, -1
    endif

; obtain information about the dimensions of the divisions
;
    pdims = [1, 1, 1]
    dims  = [1, 1, 1]
    ng    = 0

    attr_index = hdf_sd_attrfind(sd_id, 'pdims')
    if (attr_index ne -1) then hdf_sd_attrinfo, sd_id, attr_index, data = pdims

; obtain information about chunk resolution and number of ghost layers
;
    attr_index = hdf_sd_attrfind(sd_id, 'dims')
    if (attr_index ne -1) then hdf_sd_attrinfo, sd_id, attr_index, data = dims

; obtain information about the number of boundary layers
;
    if (strmid(file,0,1) eq 'r') then begin  ; TODO: this doesn't work if file name contains a path
      attr_index = hdf_sd_attrfind(sd_id, 'ng')
      if (attr_index ne -1) then hdf_sd_attrinfo, sd_id, attr_index, data = ng
    endif

; calculate  size of the full array
;
    ng = replicate(ng, n_elements(dims))
    ii = where(dims eq 1, c)
    if (c ge 1) then ng[ii] = 0
    sz = dims - 2 * ng

; create the array
;
    tmp = fltarr(pdims*sz)

; obtain information about the number of files
;
    attr_index = hdf_sd_attrfind(sd_id, 'ncpu')
    if (attr_index eq -1) then attr_index = hdf_sd_attrfind(sd_id, 'rank')
    if (attr_index ne -1) then hdf_sd_attrinfo, sd_id, attr_index, data = ncpu

; end SD interface for the reference file
;
    hdf_sd_end, sd_id

; iterate over all files and read data
;
    pos = strpos(file, string(ncpu[0], format = '(i3.3,".hdf")'))

    for n = 0, ncpus[0]-1 do begin

; prepare the filename
;
      strput, file, string(n, format = '(i3.3)'), pos

; start the SD interface
;
      sd_id = hdf_sd_start(file, /read)

; read coordinates of sub-cube
;
      attr_index = hdf_sd_attrfind(sd_id, 'pcoords')
      if (attr_index ne -1) then hdf_sd_attrinfo, sd_id, attr_index, data = pcoords

; select correct dataset
;
      sds_id = hdf_sd_select(sd_id, sds_index)

; read data from this dataset
;
      hdf_sd_getdata, sds_id, data

; and access to this dataset
;
      hdf_sd_endaccess, sds_id

; read coordinates
;
      attr_index = hdf_sd_attrfind(sd_id, 'x')
      if (attr_index ne -1) then hdf_sd_attrinfo, sd_id, attr_index, data = x
      attr_index = hdf_sd_attrfind(sd_id, 'y')
      if (attr_index ne -1) then hdf_sd_attrinfo, sd_id, attr_index, data = y
      attr_index = hdf_sd_attrfind(sd_id, 'z')
      if (attr_index ne -1) then hdf_sd_attrinfo, sd_id, attr_index, data = z

; end SD interface
;
      hdf_sd_end, sd_id

; insert sub-cube into global array
;
      ib = ng[0]
      ie = ib + sz[0] - 1
      jb = ng[1]
      je = jb + sz[1] - 1
      kb = ng[2]
      ke = kb + sz[2] - 1

      data = temporary(data[ib:ie,jb:je,kb:ke])

      ib = pcoords[0] * sz[0]
      ie = ib + sz[0] - 1
      jb = pcoords[1] * sz[1]
      je = jb + sz[1] - 1
      kb = pcoords[2] * sz[2]
      ke = kb + sz[2] - 1

      tmp[ib:ie,jb:je,kb:ke] = data

      data = tmp
    endfor

  endelse

  return, data
end