;===============================================================================
;
;  function LOADMHDHDF5 reads scientific data produced by Godunov-MHD
;  written in HDF5 format.
;
;-------------------------------------------------------------------------------
;
;  arguments:
;
;    file - file to read data from
;    var  - variable to be readed
;
function loadmhdhdf5, file, var, xcoord = x, ycoord = y, zcoord = z, time = t, single = single, info = info, short = short, quiet = quiet

  data = -1
  ncpus = 1

; check it the file exists
;
  if (not file_test(file, /regular, /read)) then begin
    print, 'This file does not exist!'
    return, data
  endif

; check it the file is really written in HDF format
;
  if (not h5f_is_hdf5(file)) then begin
    print, 'This file is not in HDF5 format!'
    return, data
  endif

; open a file
;
  fid = h5f_open(file)

; check if there were problems with opening the SD interface
;
  if (fid le 0) then begin
    print, 'Could not open file!'

; close a file
;
    h5f_close, fid

; return -1
;
    return, -1
  endif

; everything goes fine so far, so proceed with the file
;
; if keyword /LIST is defined list all variables only
;
  if (keyword_set(info)) then begin

; open group 'attributes'
;
    gid = h5g_open(fid, 'attributes')

; get number of attributes
;
    nattributes = h5a_get_num_attrs(gid)

; print all attribute names
;
    print, 'Number of attributes:', nattributes
    attrs = strarr(nattributes)
    print, 'Attribute names:'
    for n = 0, nattributes - 1 do begin
      aid = h5a_open_idx(gid, n)
      nm = h5a_get_name(aid)
      h5a_close, aid
      attrs[n] = nm
    endfor
    print, '    ', attrs
    print, ''

; close group 'attributes'
;
    h5g_close, gid

; open group 'variables'
;
    gid = h5g_open(fid, 'variables')

; get number of datasets
;
    ndatasets = h5g_get_num_objs(gid)

; print all dataset names
;
    print, 'Number of datasets  :', ndatasets
    names = strarr(ndatasets)
    print, 'Dataset names:'
    for n = 0, ndatasets - 1 do begin
      nm = h5g_get_obj_name_by_idx(gid, n)
      names[n] = nm
    endfor
    print, '    ', names
    print, ''

; close group 'variables'
;
    h5g_close, gid

; close a file
;
    h5f_close, fid

; return 0
;
    return, 0
  endif

; open group 'attributes'
;
  gid = h5g_open(fid, 'attributes')

; read attributes
;
  t      = 0.0
  nprocs = 1

; get number of attributes
;
  nattributes = h5a_get_num_attrs(gid)

; read attributes
;
  for n = 0, nattributes - 1 do begin
    aid = h5a_open_idx(gid, n)
    name = h5a_get_name(aid)
    data = h5a_read(aid)
    case name of
      'time'   : t      = data[0]
      'nprocs' : nprocs = data[0]
    else :
    endcase
    h5a_close, aid
  endfor

; close group 'attributes'
;
  h5g_close, gid

; proceed if ncpus == 1 or keword /SINGLE is set
;
  if (nprocs eq 1 or keyword_set(single)) then begin  ; data in a single file

; open group 'coordinates'
;
    gid = h5g_open(fid, 'coordinates')

; read coordinates
;
    did = h5d_open(gid, 'xc')
    x = h5d_read(did)
    h5d_close, did
    did = h5d_open(gid, 'yc')
    y = h5d_read(did)
    h5d_close, did
    did = h5d_open(gid, 'zc')
    z = h5d_read(did)
    h5d_close, did

; close group 'coordinates'
;
    h5g_close, gid

; open group 'variables'
;
    gid = h5g_open(fid, 'variables')

; get number of datasets
;
    ndatasets = h5g_get_num_objs(gid)

; find variables
;
    for n = 0, ndatasets - 1 do begin
      name = h5g_get_obj_name_by_idx(gid, n)
      if (name eq var) then begin
        did = h5d_open(gid, name)
        data = h5d_read(did)
        h5d_close, did
      endif
    endfor

; variable not found
;
    if (n_elements(data) eq 1) then begin
      if (not keyword_set(quiet)) then begin
        print, 'Could not find variable ', var, ' in the file.'
        print, 'Please add keyword /INFO to see all stored variables.'
      endif
      h5g_close, gid
      h5f_close, fid
      return, -1
    endif

; close group 'variables'
;
    h5g_close, gid

; close a file
;
    h5f_close, fid

  endif else begin                                                    ; data in multiple files

; open group 'variables'
;
    gid = h5g_open(fid, 'variables')

; get number of datasets
;
    ndatasets = h5g_get_num_objs(gid)

; find the index of requested dataset
;
    dset_index = -1
    for n = 0, ndatasets - 1 do begin
      name = h5g_get_obj_name_by_idx(gid, n)
      if (name eq var) then begin
        dset_index = n
      endif
    endfor

; variable not found
;
    if (dset_index eq -1) then begin
      if (not keyword_set(quiet)) then begin
        print, 'Could not find variable ', var, ' in the file.'
        print, 'Please add keyword /INFO to see all stored variables.'
      endif
      h5g_close, gid
      h5f_close, fid
      return, -1
    endif

; close group 'variables'
;
    h5g_close, gid

; obtain information about the dimensions of the divisions
;
    pdims = [1, 1, 1]
    dims  = [1, 1, 1]
    ng    = 0
    nproc = 0

; open group 'attributes'
;
    gid = h5g_open(fid, 'attributes')

; read attributes
;
    for n = 0, nattributes - 1 do begin
      aid = h5a_open_idx(gid, n)
      name = h5a_get_name(aid)
      data = h5a_read(aid)
      case name of
        'pdims'  : pdims  = data
        'dims'   : dims   = data
        'ng'     : ng     = data[0]
        'nproc'  : nproc  = data[0]
      else :
      endcase
      h5a_close, aid
    endfor

; close group 'attributes'
;
    h5g_close, gid

; calculate  size of the full array
;
    ng = replicate(ng, n_elements(dims))
    ii = where(dims eq 1, c)
    if (c ge 1) then ng[ii] = 0
    sz = dims - 2 * ng

; create the array
;
    tmp = fltarr(pdims*sz)
    x   = fltarr(pdims[0]*sz[0])
    y   = fltarr(pdims[1]*sz[1])
    z   = fltarr(pdims[2]*sz[2])

; close a file
;
    h5f_close, fid

; iterate over all files and read data
;
    if (keyword_set(short)) then begin
      pos = strpos(file, string(nproc, format = '(i3.3,".h5")'))
    endif else begin
      pos = strpos(file, string(nproc, format = '(i5.5,".h5")'))
    endelse

    for p = 0, nprocs-1 do begin

; prepare the filename
;
    if (keyword_set(short)) then begin
      strput, file, string(p, format = '(i3.3)'), pos
    endif else begin
      strput, file, string(p, format = '(i5.5)'), pos
    endelse

; open a file
;
      fid = h5f_open(file)

; open group 'attributes'
;
      gid = h5g_open(fid, 'attributes')

      for n = 0, nattributes - 1 do begin
        aid = h5a_open_idx(gid, n)
        name = h5a_get_name(aid)
        data = h5a_read(aid)
        case name of
          'pcoords': pcoords  = data
        else :
        endcase
        h5a_close, aid
      endfor

; close group 'attributes'
;
      h5g_close, gid

; open group 'coordinates'
;
      gid = h5g_open(fid, 'coordinates')

; read coordinates
;
      did = h5d_open(gid, 'xc')
      xt = h5d_read(did)
      h5d_close, did
      did = h5d_open(gid, 'yc')
      yt = h5d_read(did)
      h5d_close, did
      did = h5d_open(gid, 'zc')
      zt = h5d_read(did)
      h5d_close, did

; close group 'coordinates'
;
      h5g_close, gid

; open group 'variables'
;
      gid = h5g_open(fid, 'variables')

; find variables
;
      for n = 0, ndatasets - 1 do begin
        name = h5g_get_obj_name_by_idx(gid, n)
        if (name eq var) then begin
          did = h5d_open(gid, name)
          data = h5d_read(did)
          h5d_close, did
        endif
      endfor

; close group 'variables'
;
      h5g_close, gid

; close a file
;
      h5f_close, fid

; insert sub-cube into global array
;
      ib = ng[0]
      ie = ib + sz[0] - 1
      jb = ng[1]
      je = jb + sz[1] - 1
      kb = ng[2]
      ke = kb + sz[2] - 1

      data = temporary(data[ib:ie,jb:je,kb:ke])
      xt   = temporary(xt[ib:ie])
      yt   = temporary(yt[jb:je])
      zt   = temporary(zt[kb:ke])

      ib = pcoords[0] * sz[0]
      ie = ib + sz[0] - 1
      jb = pcoords[1] * sz[1]
      je = jb + sz[1] - 1
      kb = pcoords[2] * sz[2]
      ke = kb + sz[2] - 1

      tmp[ib:ie,jb:je,kb:ke] = data

      x[ib:ie] = xt
      y[jb:je] = yt
      z[kb:ke] = zt

    endfor

    data = tmp

  endelse

  return, data
end