;===============================================================================
;
;  Subroutine READ_SLICE reads a slice file produced by Godunov
;
;  Currently supports HDF5 format only.
;
;-------------------------------------------------------------------------------
;
;  arguments:
;
;    fl  - file to read data from
;    var  - variable to be readed
;
function read_slice, fl, var, info = in, nsnapshots = ns, snapshot = sp, first_snapshot = nf, last_snapshot = nl, direction = dr, t = t, x = x, y = y, verbose = verbose

; check if file is in HDF5 format
;
  if not h5f_is_hdf5(fl) then begin
    message, fl + 'is not an HDF5 file.'
  endif

; open file and open group 'attributes'
;
  fid = h5f_open(fl)
  gid = h5g_open(fid, 'attributes')

; get number of attributes
;
  nattributes = h5a_get_num_attrs(gid)

; read attributes
;
  nprocs = 1
  nf     = 0
  nl     = 0
  ns     = 0
  dm     = [1, 1, 1]
  pm     = [1, 1, 1]
  for n = 0, nattributes - 1 do begin
    aid = h5a_open_idx(gid, n)
    name = h5a_get_name(aid)
    data = h5a_read(aid)
    case name of
      'nprocs'         : nprocs = data[0]
      'first_snapshot' : nf     = data[0]
      'last_snapshot'  : nl     = data[0]
      'nsnapshots'     : ns     = data[0]
      'dims'           : dm     = data
      'pdims'          : pm     = data
    else :
    endcase
    h5a_close, aid
  endfor
  ns = max([ns, nl - nf + 1])


; compute total dimensions
;
  fm = dm * pm

; close this attribute and group 'attributes'
;
  h5g_close, gid

; close file
;
  h5f_close, fid

  if keyword_set(in) then begin                  ; print information about stored data

; print info
;
    print, 'Filename           : ', fl
    print, 'Number of frames   : ', ns
    print, 'First frame        : ', nf
    print, 'Last frame         : ', nl
    print, 'Dimensions         : ', dm
    print, 'Division           : ', pm

  endif

  qty = -1

; read particles
;
  if (n_elements(sp) eq 1 and keyword_set(dr)) then begin                  ; read a given snapshot
    if (keyword_set(verbose)) then $
      print, 'reading ' + var + ' for snapshot : ' + string(sp) + ' [' + strupcase(string(dr)) + ' plane]'

    if (sp lt 0 or sp gt nl) then message, 'Snapshot out of range!'

; prepare output array
;
    case strupcase(string(dr)) of
    'XY' : begin
             qty = fltarr(fm[0], fm[1])
             x   = fltarr(fm[0])
             y   = fltarr(fm[1])
           end
    'XZ' : begin
             qty = fltarr(fm[0], fm[2])
             x   = fltarr(fm[0])
             y   = fltarr(fm[2])
           end
    'YZ' : begin
             qty = fltarr(fm[1], fm[2])
             x   = fltarr(fm[1])
             y   = fltarr(fm[2])
           end
    else: print, 'Slice direction ' + string(dr) + ' is not implemented.'
    endcase

; iterate over all subfiles
;
    for p = 0, nprocs - 1 do begin

; generate filename
;
      fn = strmid(fl, 0, strpos(fl, '.h5')-6) + '_' + string(p, format = "(i5.5)") + '.h5'

; open file
;
      fid = h5f_open(fn)

; read attributes
;
      gid = h5g_open(fid, 'attributes')
      aid = h5a_open_name(gid, 'pcoords')
      pcoords = (h5a_read(aid))
      h5a_close, aid
      h5g_close, gid

; read coordinates
;
      gid = h5g_open(fid, 'coordinates')
      aid = h5a_open_name(gid, 'x')
      xt = (h5a_read(aid))
      h5a_close, aid
      aid = h5a_open_name(gid, 'y')
      yt = (h5a_read(aid))
      h5a_close, aid
      aid = h5a_open_name(gid, 'z')
      zt = (h5a_read(aid))
      h5a_close, aid
      h5g_close, gid

; generate snapshot name
;
      sn = 'sl' + string(sp, format="(i6.6)")

; open snapshot group
;
      sid = h5g_open(fid, sn)

; read time of this snapshot
;
      aid = h5a_open_name(sid, 'time')
      t   = (h5a_read(aid))[0]
      h5a_close, aid

; open group for requested slice direction
;
      gid = h5g_open(sid, strupcase(string(dr)))

; get number of stored variables
;
      ndatasets = h5g_get_num_objs(gid)

      if (ndatasets gt 0) then begin

; find variables
;
        for n = 0, ndatasets - 1 do begin
          name = h5g_get_obj_name_by_idx(gid, n)
          if (name eq var) then begin
            did = h5d_open(gid, name)
            data = h5d_read(did)
            h5d_close, did
          endif
        endfor

; variable not found
;
        if (n_elements(data) eq 1) then begin
          print, 'Could not find variable ', var, ' in the file.'
          print, 'Please add keyword /INFO to see all stored variables.'
        endif else begin

; insert subslice
;
          case strupcase(string(dr)) of
          'XY' : begin
                   ib = pcoords[0] * dm[0]
                   ie = ib + dm[0] - 1
                   jb = pcoords[1] * dm[1]
                   je = jb + dm[1] - 1

                   qty[ib:ie,jb:je] = data
                   x[ib:ie] = xt
                   y[jb:je] = yt
                 end
          'XZ' : begin
                   ib = pcoords[0] * dm[0]
                   ie = ib + dm[0] - 1
                   jb = pcoords[2] * dm[2]
                   je = jb + dm[2] - 1

                   qty[ib:ie,jb:je] = data
                   x[ib:ie] = xt
                   y[jb:je] = zt
                 end
          'YZ' : begin
                   ib = pcoords[1] * dm[1]
                   ie = ib + dm[1] - 1
                   jb = pcoords[2] * dm[2]
                   je = jb + dm[2] - 1

                   qty[ib:ie,jb:je] = data
                   x[ib:ie] = yt
                   y[jb:je] = zt
                 end
          else: print, 'Slice direction ' + string(dr) + ' is not implemented.'
          endcase

        endelse

      endif

; close group
;
      h5g_close, gid

; close group
;
      h5g_close, sid

; close file
;
      h5f_close, fid

    endfor

  endif

  return, qty
end
