!!******************************************************************************
!!
!!  This file is part of the GODUNOV source code, a program to perform
!!  Newtonian or relativistic magnetohydrodynamical simulations.
!!
!!  Copyright (C) 2007-2013 Grzegorz Kowal <grzegorz@amuncode.org>
!!
!!  This program is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  You should have received a copy of the GNU General Public License
!!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
!!
!!******************************************************************************
!!
!! module: ALGEBRA
!!
!!  This module provides all sort of algebra subroutines.
!!
!!******************************************************************************
!
module algebra

! module variables are not implicit by default
!
  implicit none

! module variables for function RPOLY and its subroutines
!
  integer                        , save :: n, nn
  real                           , save :: sr, si, u, v, a, b, c, d
  real                           , save :: a1, a2, a3, a6, a7
  real                           , save :: e, f, g, h, szr, szi, lzr, lzi
  real                           , save :: eta, are, mre
  real, dimension(:), allocatable, save :: p, qp, k, qk, svk

! interfaces
!
  interface poly_term
    module procedure poly_term_rr, poly_term_cc
  end interface
  interface poly
    module procedure poly_rr, poly_cc
  end interface

! by default everything is private
!
  private

! declare public subroutines
!
  public :: quadratic, cubic, quartic, quartic_zroots, poly

!- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
  contains
!
!===============================================================================
!
! function QUADRATIC:
! ------------------
!
!   Function finds real roots of a quadratic equation:
!
!     f(x) = a₂ x² + a₁ x + a₀ = 0
!
!   Remark:
!
!     The coefficients are renormalized in order to avoid overflow and
!     unnecessary underflow.  The catastrophic cancellation is avoided as well.
!
!   Arguments:
!
!     a₂, a₁, a₀ - the quadratic equation coefficients;
!     x          - the root array;
!
!   Return value:
!
!     The number of roots found.
!
!===============================================================================
!
  function quadratic(a, x) result(nr)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    real, dimension(3), intent(in)  :: a
    real, dimension(2), intent(out) :: x
    integer                         :: nr

! local variables
!
    real, dimension(3) :: b
    real               :: dl, dr, tm
!
!-------------------------------------------------------------------------------
!
! in order to avoid overflow or needless underflow, divide all coefficients by
! the maximum among them
!
    tm = maxval(abs(a(1:3)))
    b(1:3) = a(1:3) / tm

! check if the coefficient a₂ is not zero
!
    if (b(3) /= 0.0d+00) then

! check if the coefficient a₀ is not zero
!
      if (b(1) /= 0.0d+00) then

! coefficients a₂ and a₀ are nonzero, so we solve the quadratic equation
!
!   a₂ x² + a₁ x + a₀ = 0
!
! calculate Δ = a₁² - 4 a₂ a₀
!
        dl = b(2) * b(2) - 4.0d+00 * b(3) * b(1)

! check if Δ is larger then zero
!
        if (dl > 0.0d+00) then

! calculate quare root of Δ
!
          dr = sqrt(dl)

! Δ > 0, so the quadratic has two real roots
!
          if (b(2) > 0.0d+00) then
            tm   = 0.5d+00 * (b(2) + dr)
            x(1) = - tm / b(3)
            x(2) = - b(1) / tm
          else if (b(2) < 0.0d+00) then
            tm   = 0.5d+00 * (b(2) - dr)
            x(1) = - b(1) / tm
            x(2) = - tm / b(3)
          else
            x(2) = 0.5d+00 * dr / b(3)
            x(1) = - x(2)
          end if

! update the number of roots
!
          nr   = 2

        else if (dl == 0.0d+00) then

! Δ = 0, so the quadratic has two identical real roots
!
          x(:) = - 0.5d+00 * b(2) / b(3)

! update the number of roots
!
          nr   = 2

        else

! Δ < 0, so the quadratic does not have any real roots
!
          x(1) = 0.0d+00
          x(2) = 0.0d+00

! update the number of roots
!
          nr   = 0

        end if

      else

! since a₀ is zero, we have one zero root and the second root is calculated from
! the linear formula
!
!   (a₂ x + a₁) x = 0
!
        x(1) = 0.0d+00
        x(2) = - b(2) / b(3)

! update the number of roots
!
        nr   = 2

      end if

    else

! since coefficient a₂ is zero, the quadratic equation is reduced to a linear
! one; one root is undetermined in this case
!
!   a₁ x + a₀ = 0
!
      x(2) = 0.0d+00

! find the unique root
!
      if (b(2) /= 0.0d+00) then

! the coefficient a₁ is not zero, so the quadratic has only one root
!
        x(1) = - b(1) / b(2)

! update the number of roots
!
        nr   = 1

      else

! the coefficient a₁ is zero, so the quadratic has no roots
!
!   a₀ = 0
!
        x(1) = 0.0d+00

! update the number of roots
!
        nr   = 0

      end if

    end if

!-------------------------------------------------------------------------------
!
  end function quadratic
!
!===============================================================================
!
! function CUBIC:
! --------------
!
!   Function finds real roots of a cubic equation:
!
!     f(x) = a₃ x³ + a₂ x² + a₁ x + a₀ = 0
!
!   Remark:
!
!     The coefficients are renormalized in order to avoid overflow and
!     unnecessary underflow.
!
!   Arguments:
!
!     a(:) - the cubic equation coefficients vector (in increasing moments);
!     x(:) - the root vector;
!
!   Return value:
!
!     The number of roots found.
!
!===============================================================================
!
  function cubic(a, x) result(nr)

! include external constants
!
    use constants, only : pi2

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    real, dimension(4), intent(in)  :: a
    real, dimension(3), intent(out) :: x
    integer                         :: nr

! local variables
!
    integer :: p, n, m
    real    :: ac, bc, qc, cc, rc, th, cs

! local parameters
!
    real, parameter :: c0 = 1.0d+00 / 3.0d+00
!
!-------------------------------------------------------------------------------
!
! check if the coefficient a₃ is not zero
!
    if (a(4) /= 0.0d+00) then

! check if the coefficient a₀ is not zero
!
      if (a(1) /= 0.0d+00) then

! both coefficients a₃ and a₀ are not zero, so we have to solve full cubic
! equation
!
!   a₃ x³ + a₂ x² + a₁ x + a₀ = 0
!
! compute coefficient A = 2 a₂³ - 9 a₃ a₂ a₁ + 27 a₃² a₀
!
        ac = 2.0d+00 * a(3)**3 - 9.0d+00 * a(4) * a(3) * a(2)                  &
                                                   + 27.0d+00 * a(4)**2 * a(1)

! compute coefficient B = a₂² - 3 a₃ a₁
!
        bc = a(3)**2 - 3.0d+00 * a(4) * a(2)

! compute coefficient Q² = A² - 4 B³
!
        qc = ac**2 - 4.0d+00 * bc**3

! check the condition for any root existance
!
        if (qc > 0.0d+00) then

! the case Q > 0, in which only one root is real
!
! calculate C³ = ½ (Q + A)
!
          cc = (0.5d+00 * (sqrt(qc) + ac))**c0

! calculate the unique real root
!
          x(1) = - c0 * (a(3) + cc + bc / cc) / a(4)
          x(2) = 0.0d+00
          x(3) = 0.0d+00

! update the number of roots
!
          nr = 1

        else if (qc == 0.0d+00) then

          if (bc /= 0.0d+00) then

! the case Q = 0 and B ≠ 0, in which all three roots are real and one root
! is double
!
            x(1) = 0.5d+00 * (9.0d+00 * a(4) * a(1) - a(3) * a(2)) / bc
            x(2) = x(1)
            x(3) = (- 9.0d+00 * a(4) * a(1) + 4.0d+00 * a(3) * a(2)            &
                                                        - a(3)**3 / a(4)) / bc

! update the number of roots
!
            nr = 3

          else

! the case Q = 0 and B = 0, in which all three roots are real and equal
!
            x(:) = - c0 * a(3) / a(4)

! update the number of roots
!
            nr = 3

          end if

        else

! Q < 0, so there are three distinct and real roots
!
! compute Q and R
!
          th = 3.0d+00 * a(4)
          qc = bc / th**2
          rc = 0.5d+00 * ac / th**3

! compute cosine
!
          cs = sign(0.5d+00, ac * bc) * sqrt(ac**2 / abs(bc)**3)

! check condition
!
          if (cs <= 1.0d+00) then

! there are three real roots; prepare coefficients to calculate them
!
            th = acos(cs)
            ac = - 2.0d+00 * sqrt(qc)
            bc = - c0 * (a(3) / a(4))

! calculate roots
!
            x(1) = ac * cos(c0 *  th       ) + bc
            x(2) = ac * cos(c0 * (th + pi2)) + bc
            x(3) = ac * cos(c0 * (th - pi2)) + bc

! update the number of roots
!
            nr = 3

          else

! there is only one real root; prepare coefficients to calculate it
!
            ac = - sign(1.0d+00, rc) * (abs(rc) + sqrt(rc**2 - qc**3))**c0
            if (ac /= 0.0d+00) then
              bc = qc / ac
            else
              bc = 0.0d+00
           end if

! calculate the root
!
            x(1) = (ac + bc) - c0 * (a(3) / a(4))

! reset remaining roots
!
            x(2) = 0.0d+00
            x(3) = 0.0d+00

! update the number of roots
!
            nr = 1

          end if

        end if

      else

! since the coefficient a₀ is zero, there is one zero root, and the remaining
! roots are found from a quadratic equation
!
!   (a₃ x² + a₂ x + a₁) x = 0
!
        x(3) = 0.0d+00

! call function quadratic() to find the remaining roots
!
        nr = quadratic(a(2:4), x(1:2))

! increase the number of roots
!
        nr = nr + 1

      end if

    else

! since the coefficient a₃ is zero, the equation reduces to a quadratic one,
! and one root is undetermined
!
!   a₂ x² + a₁ x + a₀ = 0
!
      x(3) = 0.0d+00

! call function quadratic() to find the remaining roots
!
      nr = quadratic(a(1:3), x(1:2))

    end if

! sort roots in the decreasing order by their real parts
!
    do p = 2, nr
      n = p
      do while(n > 1)
        m = n - 1
        if (x(n) > x(m)) then
          qc   = x(m)
          x(m) = x(n)
          x(n) = qc
        end if
        n = m
      end do
    end do

!-------------------------------------------------------------------------------
!
  end function cubic
!
!===============================================================================
!
! subroutine QUARTIC:
! ------------------
!
!   Subroutine finds real roots of a quartic equation:
!
!     f(x) = a₄ * x⁴ + a₃ * x³ + a₂ * x² + a₁ * x + a₀ = 0
!
!   Arguments:
!
!     a(:) - the quartic equation coefficients vector (in increasing moments);
!     x(:) - the root vector;
!
!   References:
!
!     Hacke, Jr. J. E., "A Simple Solution of the General Quartic",
!     The American Mathematical Monthly, 1941, vol. 48, no. 5, pp. 327-328
!
!===============================================================================
!
  function quartic(a, x) result(nr)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    real, dimension(5), intent(in)  :: a
    real, dimension(4), intent(out) :: x
    integer                         :: nr

! local variables
!
    integer            :: p, n, m
    real               :: pc, qc, rc
    real               :: zm, k2, l2, kc, lc

! local arrays
!
    real, dimension(4) :: b
    real, dimension(2) :: y
    real, dimension(3) :: c, z
!
!-------------------------------------------------------------------------------
!
! check if the coefficient a₄ is not zero
!
    if (a(5) /= 0.0d+00) then

! check if the coefficient a₀ is not zero
!
      if (a(1) /= 0.0d+00) then

! both coefficients a₄ and a₀ are not zero, so we have to solve full quartic
! equation
!
!   a₄ * x⁴ + a₃ * x³ + a₂ * x² + a₁ * x + a₀ = 0
!
! calculate coefficients of P, Q, and R
!
        pc = (- 3.0d+00 * a(4)**2 + 8.0d+00 * a(5) * a(3)) / (8.0d+00 * a(5)**2)
        qc = (a(4)**3 - 4.0d+00 * a(5) * a(4) * a(3)                           &
                             + 8.0d+00 * a(5)**2 * a(2)) / (8.0d+00 * a(5)**3)
        rc = (- 3.0d+00 * a(4)**4 + 16.0d+00 * a(5) * a(4)**2 * a(3)           &
              - 64.0d+00 * a(5)**2 * a(4) * a(2) + 256.0d+00 * a(5)**3 * a(1)) &
                                                       / (256.0d+00 * a(5)**4)

! prepare coefficients for the cubic equation
!
        b(4) =   8.0d+00
        b(3) = - 4.0d+00 * pc
        b(2) = - 8.0d+00 * rc
        b(1) =   4.0d+00 * pc * rc - qc**2

! find roots of the cubic equation
!
        nr = cubic(b(1:4), z(1:3))

! check if the cubic solver returned any roots
!
        if (nr > 0) then

! take the maximum root
!
          zm = z(1)

! calculate coefficients to find quartic roots
!
          k2 = 2.0d+00 * zm - pc
          l2 =      zm * zm - rc

! check if both coefficients are larger then zero
!
          if (k2 > 0.0d+00 .or. l2 > 0.0d+00) then

! prepare coefficients K and L
!
            if (k2 > l2) then
              kc = sqrt(k2)
              lc = 0.5d+00 * qc / kc
            else if (l2 > k2) then
              lc = sqrt(l2)
              kc = 0.5d+00 * qc / lc
            else
              kc = sqrt(k2)
              lc = sqrt(l2)
            end if

! prepare coefficients for the first quadratic equation
!
            c(3) = 1.0d+00
            c(2) = - kc
            c(1) = zm + lc

! find the first pair of roots
!
            m = quadratic(c(1:3), y(1:2))

! update the roots
!
            do n = 1, m
              x(n) = y(n) - 0.25d+00 * a(4) / a(5)
            end do

! update the number of roots
!
            nr = m

! prepare coefficients for the second quadratic equation
!
            c(3) = 1.0d+00
            c(2) = kc
            c(1) = zm - lc

! find the second pair of roots
!
            m = quadratic(c(1:3), y(1:2))

! update the roots
!
            do n = 1, m
              x(nr + n) = y(n) - 0.25d+00 * a(4) / a(5)
            end do

! update the number of roots
!
            nr = nr + m

! reset the remainings roots
!
            do n = nr + 1, 4
              x(n) = 0.0d+00
            end do

          else

! both K² and L² are negative, which signifies that the quartic equation has
! no real roots; reset them
!
            x(:) = 0.0d+00

! update the number of roots
!
            nr   = 0

          end if

        else

! weird, no cubic roots... it might mean that this is not real quartic equation
! or something is wrong; reset the roots
!
          x(:) = 0.0d+00

        end if

      else

! since the coefficient a₀ is zero, there is one zero root, and the remaining
! roots are found from a cubic equation
!
!   (a₄ * x³ + a₃ * x² + a₂ * x + a₁) * x = 0
!
        x(4) = 0.0d+00

! call function cubic() to find the remaining roots
!
        nr = cubic(a(2:5), x(1:3))

! increase the number of roots
!
        nr = nr + 1

      end if

    else

! since the coefficient a₄ is zero, the equation reduces to a cubic one, and
! one root is undetermined
!
!   a₃ * x³ + a₂ * x² + a₁ * x + a₀ = 0
!
      x(4) = 0.0d+00

! call function cubic() to find the remaining roots
!
      nr = cubic(a(1:4), x(1:3))

    end if

! sort roots in the decreasing order by their real parts
!
    do p = 2, nr
      n = p
      do while(n > 1)
        m = n - 1
        if (x(n) > x(m)) then
          qc   = x(m)
          x(m) = x(n)
          x(n) = qc
        end if
        n = m
      end do
    end do

!-------------------------------------------------------------------------------
!
  end function quartic
!
!===============================================================================
!
! subroutine QUARTIC_ZROOTS:
! -------------------------
!
!   Subroutine finds real roots of a quartic equation:
!
!     f(x) = a₄ * x⁴ + a₃ * x³ + a₂ * x² + a₁ * x + a₀ = 0
!
!   Arguments:
!
!     a(:) - the quartic equation coefficients vector (in increasing moments);
!     x(:) - the root vector;
!
!   References:
!
!     Hacke, Jr. J. E., "A Simple Solution of the General Quartic",
!     The American Mathematical Monthly, 1941, vol. 48, no. 5, pp. 327-328
!
!===============================================================================
!
  function quartic_zroots(a, x) result(nr)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    real, dimension(5), intent(in)  :: a
    real, dimension(4), intent(out) :: x
    integer                         :: nr

! local variables
!
    integer :: p, m, n
    real    :: t

! local arrays
!
    complex, dimension(5)   :: b
    complex, dimension(4)   :: z
!
!-------------------------------------------------------------------------------
!
! copy the coefficients to vector B and normalize by the largest coefficient
!
    b(:) = cmplx(a(:)) / maxval(abs(a(:)))

! use subroutine zroots to find all complex roots
!
    call zroots(b(:), z(:))

! check all roots if they are real
!
    nr = 0
    p  = 4
    do while(p > 0)
      if (aimag(z(p)) == 0.0d+00) then
        nr = nr + 1
        x(nr) = real(z(p))
      end if
      p = p - 1
    end do

! sort roots in the decreasing order by their real parts
!
    do p = 2, nr
      n = p
      do while(n > 1)
        m = n - 1
        if (x(n) > x(m)) then
          t    = x(m)
          x(m) = x(n)
          x(n) = t
        end if
        n = m
      end do
    end do

!-------------------------------------------------------------------------------
!
  end function quartic_zroots
!
!===============================================================================
!
! subroutine ZROOTS:
! -----------------
!
!   Subroutine finds all complex roots of the complex polynomial:
!
!     f(x) = aₙ * xⁿ + ... + a₃ * x³ + a₂ * x² + a₁ * x + a₀ = 0
!
!   Arguments:
!
!     a(n+1)   - the complex polynomial coefficients;
!     roots(n) - the array of complex roots;
!     polish   - flag controlling the optional polishing;
!
!   Description:
!
!     Given the array of N+1 complex coefficients 'a' of the polynomial f(x),
!     this routine successively calls laguerre() and finds all N complex roots.
!     The logical variable 'polish' should be input as .true. if polishing
!     (also by Laguerre's method) is desired, .false. if the roots will be
!     subsequently polished by other means.
!
!   References:
!
!     Press, W. H, Teukolsky, S. A., Vetterling, W. T., Flannery, B. P.,
!     "Numerical Recipes in Fortran",
!     Cambridge University Press, Cambridge, 1992
!
!===============================================================================
!
  subroutine zroots(a, x)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    complex, dimension(:), intent(in)  :: a
    complex, dimension(:), intent(out) :: x

! local variables
!
    integer :: m, n, p
    complex :: z

! local arrays
!
    complex, dimension(size(a)) :: b

! local parameters
!
    real, parameter :: eps = 1.0d-06
!
!-------------------------------------------------------------------------------
!
! get the number of expected roots
!
    n = size(a) - 1

! check if the root array has correct dimension
!
    if (size(x) /= n) then
      write(*,*) 'ZROOTS: wrong dimension of the root array!'
      stop
    end if

! copy coefficients to new array for successive deflation
!
    b(:) = a(:)

! iterate over each root to be found
!
    do p = n, 1, -1

! start at zero to favor convergence to smallest remaining root
!
      z = cmplx(0.0d+00, 0.0d+00)

! find the root
!
      call laguerre(b(1:p+1), z)

! check the imaginary part of the root
!
      if (abs(aimag(z)) <= eps * abs(real(z))) z = cmplx(real(z), 0.0d+00)

! substitute the root to output array
!
      x(p) = z

! forward deflation
!
      b(p:1:-1) = poly_term(b(p+1:2:-1), z)

    end do

!-------------------------------------------------------------------------------
!
  end subroutine zroots
!
!===============================================================================
!
! subroutine LAGUERRE:
! -------------------
!
!   Given the array of N+1 complex coefficients 'a' of the polynomial
!
!     f(x) = aₙ * xⁿ + ... + a₃ * x³ + a₂ * x² + a₁ * x + a₀ = 0,
!
!   this routine successively calls laguerre() and finds all N complex roots.
!   The logical variable 'polish' should be input as .true. if polishing
!   (also by Laguerre's method) is desired, .false. if the roots will be
!   subsequently polished by other means.
!
!   Arguments:
!
!     a(n+1) - the complex polynomial coefficients;
!     x      - the complex root;
!     ni     - the iteration number;
!
!   References:
!
!     Press, W. H, Teukolsky, S. A., Vetterling, W. T., Flannery, B. P.,
!     "Numerical Recipes in Fortran",
!     Cambridge University Press, Cambridge, 1992
!
!===============================================================================
!
  subroutine laguerre(a, z)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    complex, dimension(:), intent(in)    :: a
    complex              , intent(inout) :: z

! local variables
!
    integer  :: n, m
    real     :: x, ap, am, er
    complex  :: dz, zn, f, g, h, sq, gp, gm, g2

! local arrays
!
    complex, dimension(size(a)) :: b, d

! local parameters
!
    integer               , parameter :: mr = 8, mt = 10, nmax = mt * mr
    real                  , parameter :: eps = 2.0d+00 * epsilon(1.0d+00)
    real   , dimension(mr), parameter :: frac = (/ 5.0d-01, 2.5d-01, 7.5d-01   &
                                                 , 1.3d-01, 3.8d-01, 6.2d-01   &
                                                 , 8.8d-01, 1.0d+00 /)
!
!-------------------------------------------------------------------------------
!
! get the number of expected roots
!
    m = size(a) - 1

! loop over iterations up to allowed maximum
!
    do n = 1, nmax

! store the absolute value of the root
!
      x = abs(z)

! efficient computation of the polynomial and its first two derivatives
!
      b(m+1:1:-1) = poly_term(a(m+1:1:-1), z)
      d(m  :1:-1) = poly_term(b(m+1:2:-1), z)
      f           = poly     (d(2:m)     , z)

! estimate of roundoff in evaluating polynomial
!
      er = max(1.0d-64, eps * poly(abs(b(1:m+1)), x))

! the root has been found
!
      if (abs(b(1)) <= er) return

! the generic case, use Laguerre's formula
!
      g  = d(1) / b(1)
      g2 = g * g
      h  = g2 - 2.0d+00 * f / b(1)
      sq = sqrt((m - 1) * (m * h - g2))
      gp = g + sq
      gm = g - sq
      ap = abs(gp)
      am = abs(gm)
      if (ap < am) gp = gm
      if (max(ap,am) > 0.0d+00) then
        dz = m / gp
      else
        dz = exp(cmplx(log(1.0d+00 + x), n, kind = 8))
      end if
      zn  = z - dz

! the root converged
!
      if (z == zn) return

! every so often we take a fraction step, to break any limit cycle (itself
! a rare occurance)
!
      if (mod(n, mt) /= 0) then
        z = zn
      else
        z = z - dz * frac(n / mt)
      end if

    end do

#ifdef DEBUG
! print warning, since the maximum number of iterations has been reached
!
    write (*,*)
    write (*,"(1x,a)") "LAGUERRE: Too many iterations!"
    write (*,"(1x,5(1pe25.16))") a(5:1:-1)
#endif /* DEBUG */

!-------------------------------------------------------------------------------
!
  end subroutine laguerre
!
!===============================================================================
!
! function POLY_TERM_RR:
! ---------------------
!
!   Returns an array of type and size the same as the one-dimensional array A,
!   containing the partial cumulants of the polynomial with coefficients A
!   (arranged from highest-order to lowest-order coefficients, n.b.) evaluated
!   at x.  This is equivalent to synthetic division, and can be parallelized.
!   See §22.3. Note that the order of arguments is reversed in poly and poly
!   term — each routine returns a value with the size and shape of the first
!   argument, the usual Fortran 90 convention.
!
!   Arguments:
!
!     a(:) - the polynomial coefficients;
!     x    - the value for which the polynomia is evaluated;
!
!   References:
!
!     Press, W. H, Teukolsky, S. A., Vetterling, W. T., Flannery, B. P.,
!     "Numerical Recipes in Fortran",
!     Cambridge University Press, Cambridge, 1992
!
!===============================================================================
!
  function poly_term_rr(a, x) result(u)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    real, dimension(:), intent(in) :: a
    real              , intent(in) :: x

! local variables
!
    integer                        :: n, i

! local arrays
!
    real, dimension(size(a))       :: u
!
!-------------------------------------------------------------------------------
!
! obtain the polynomial order
!
    n = size(a)

! exit, if there are no input coefficients
!
    if (n <= 0) return

! calculate partial cumulants
!
    u(1) = a(1)

    do i = 2, n
      u(i) = a(i) + x * u(i - 1)
    end do

!-------------------------------------------------------------------------------
!
  end function poly_term_rr
!
!===============================================================================
!
! function POLY_TERM_CC:
! ---------------------
!
!   Returns an array of type and size the same as the one-dimensional array A,
!   containing the partial cumulants of the polynomial with coefficients A
!   (arranged from highest-order to lowest-order coefficients, n.b.) evaluated
!   at x.  This is equivalent to synthetic division, and can be parallelized.
!   See §22.3. Note that the order of arguments is reversed in poly and poly
!   term — each routine returns a value with the size and shape of the first
!   argument, the usual Fortran 90 convention.
!
!   Arguments:
!
!     a(:) - the polynomial coefficients;
!     x    - the value for which the polynomia is evaluated;
!
!   References:
!
!     Press, W. H, Teukolsky, S. A., Vetterling, W. T., Flannery, B. P.,
!     "Numerical Recipes in Fortran",
!     Cambridge University Press, Cambridge, 1992
!
!===============================================================================
!
  function poly_term_cc(a, x) result(u)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    complex, dimension(:), intent(in) :: a
    complex              , intent(in) :: x

! local variables
!
    integer                           :: n, i

! local arrays
!
    complex, dimension(size(a))       :: u
!
!-------------------------------------------------------------------------------
!
! obtain the polynomial order
!
    n = size(a)

! exit, if there are no input coefficients
!
    if (n <= 0) return

! calculate partial cumulants
!
    u(1) = a(1)

    do i = 2, n
      u(i) = a(i) + x * u(i - 1)
    end do

!-------------------------------------------------------------------------------
!
  end function poly_term_cc
!
!===============================================================================
!
! function POLY_RR:
! ----------------
!
!   Evaluate a real polynomial f(x) for a real value x.
!
!   Arguments:
!
!     a(:) - the polynomial coefficients;
!     x    - the value for which the polynomial is evaluated;
!
!   References:
!
!     Press, W. H, Teukolsky, S. A., Vetterling, W. T., Flannery, B. P.,
!     "Numerical Recipes in Fortran",
!     Cambridge University Press, Cambridge, 1992
!
!===============================================================================
!
  function poly_rr(a, x) result(f)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    real, dimension(:), intent(in) :: a
    real              , intent(in) :: x
    real                           :: f

! local variables
!
    integer :: n, i
!
!-------------------------------------------------------------------------------
!
! obtain the polynomial order
!
    n = size(a)

! evaluate the polynomial value at x
!
    f = a(n)

    do i = n - 1, 1, -1
      f = x * f + a(i)
    end do

!-------------------------------------------------------------------------------
!
  end function poly_rr
!
!===============================================================================
!
! function POLY_CC:
! ----------------
!
!   Evaluate a complex polynomial f(x) for a complex value x.
!
!   Arguments:
!
!     a(:) - the polynomial coefficients;
!     x    - the value for which the polynomial is evaluated;
!
!   References:
!
!     Press, W. H, Teukolsky, S. A., Vetterling, W. T., Flannery, B. P.,
!     "Numerical Recipes in Fortran",
!     Cambridge University Press, Cambridge, 1992
!
!===============================================================================
!
  function poly_cc(a, x) result(f)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    complex, dimension(:), intent(in) :: a
    complex              , intent(in) :: x
    complex                           :: f

! local variables
!
    integer :: n, i
!
!-------------------------------------------------------------------------------
!
! obtain the polynomial order
!
    n = size(a)

! evaluate the polynomial value at x
!
    f = a(n)

    do i = n - 1, 1, -1
      f = x * f + a(i)
    end do

!-------------------------------------------------------------------------------
!
  end function poly_cc
!
!===============================================================================
!
! subroutine RPOLY:
! ----------------
!
!   Subroutine finds zeros of a real polynomial:
!
!     f(x) = aₙ * xⁿ + ... + a₃ * x³ + a₂ * x² + a₁ * x + a₀ = 0
!
!   Remarks:
!
!     To change the size of polynomials which can be solved, reset the
!     dimensions of the arrays in the common area and in the following
!     declarations. The subroutine uses single precision calculations for
!     scaling, bounds and error calculations.  All calculations for the
!     iterations are done in double precision.
!
!   Arguments:
!
!     degree       - integer degree of polynomial;
!     op           - double precision vector of coefficients in order of
!                    decreasing powers;
!     zeror, zeroi - output double precision vectors of real and imaginary parts
!                    of the zeros.
!     fail         - output logical parameter, true only if leading coefficient
!                    is zero or if rpoly has found fewer than degree zeros. In
!                    the latter case degree is reset to the number of zeros found.
!
!   References:
!
!     CACM Algorithm 493 by Jenkins & Traub, netlib Sat Jul 26 11:57:43 EDT 1986
!     Code converted using TO_F90 by Alan Miller, Date: 2003-06-02  Time: 10:42:22
!
!===============================================================================
!
  function rpoly(degree, op, zeror, zeroi) result(fail)

! include external constants
!
    use constants, only : d2r

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer                  , intent(inout) :: degree
    real, dimension(degree+1), intent(in)    :: op
    real, dimension(degree+1), intent(out)   :: zeror, zeroi
    logical                                  :: fail

! local variables
!
    logical :: zerok
    integer :: cnt, nz, i, j, jj, l, nm1
    real    :: t, aa, bb, cc, factor
    real    :: lo, mx, mn, xx, yy, cosr, sinr, xxx, x, sc, bnd, xm
    real    :: ff, df, dx, infin, smalno, base

! allocatable arrays
!
    real, dimension(:), allocatable   :: temp, pt
!
!-------------------------------------------------------------------------------
!
    base   = radix(0.0d+00)
    eta    = epsilon(1.0d+00)
    infin  = huge(0.0d+00)
    smalno = tiny(0.0d+00)

! are and mre refer to the unit error in + and * respectively, they are assumed
! to be the same as eta
!
    are = eta
    mre = eta
    lo = smalno / eta

! initialization of constants for shift rotation
!
    xx   = sqrt(0.5)
    yy   = -xx
    cosr = -0.0697564740d+00
    sinr =  0.9975640500d+00
    fail = .false.
    n    = degree
    nn   = n + 1

! algorithm fails if the leading coefficient is zero
!
    if (op(1) == 0.0d+00) then
      fail   = .true.
      return
    end if

! remove the zeros at the origin if any
!
10  if (op(nn) == 0.0d+00) then
      j        = degree - n + 1
      zeror(j) = 0.d+00
      zeroi(j) = 0.d+00
      nn       = nn - 1
      n        = n  - 1
      go to 10
    end if

! allocate various arrays
!
    if (allocated(p)) deallocate(p, qp, k, qk, svk)
    allocate(p(nn), qp(nn), k(nn), qk(nn), svk(nn), temp(nn), pt(nn))

! make a copy of the coefficients
!
    p(1:nn) = op(1:nn)

! start the algorithm for one zero
!
30  if (n <= 2) then
      if (n <  1) return

! calculate the final zero or pair of zeros
!
      if (n /= 2) then
        zeror(degree) = - p(2) / p(1)
        zeroi(degree) = 0.0d+00
        return
      end if
      call quad(p(1), p(2), p(3), zeror(degree - 1), zeroi(degree - 1),        &
                                                 zeror(degree), zeroi(degree))
      return

    end if

! find largest and smallest moduli of coefficients.
!
    mx = 0.0d+00
    mn = infin
    do i = 1, nn
      x = abs(real(p(i)))
      if (x > mx) mx = x
      if (x /= 0.0d+00 .and. x < mn) mn = x
    end do

! scale if there are large or very small coefficients computes a scale factor
! to multiply the coefficients of the polynomial; the scaling is done to avoid
! overflow and to avoid undetected underflow interfering with the convergence
! criterion; the factor is a power of the base
!
    sc = lo / mn
    if (sc <= 1.0d+00) then
      if (mx < 1.0d+01) go to 60
      if (sc == 0.0d+00) sc = smalno
    else
      if (infin / sc < mx) go to 60
    end if
    l = log(sc) / log(base) + 0.5d+00
    factor = (1.0d+00*base)**l
    if (factor /= 1.0d+00) then
      p(1:nn) = factor * p(1:nn)
    end if

! compute lower bound on moduli of zeros
!
60  pt(1:nn) = abs(p(1:nn))
    pt(nn)   = -pt(nn)

! compute upper estimate of bound
!
    x = exp((log(-pt(nn)) - log(pt(1))) / n)

! if newton step at the origin is better, use it
!
    if (pt(n) /= 0.0d+00) then
      xm = - pt(nn) / pt(n)
      if (xm < x) x = xm
    end if

! chop the interval (0,x) until ff .le. 0
!
80  xm = x * 0.1d+00
    ff = pt(1)
    do i = 2, nn
      ff = ff * xm + pt(i)
    end do
    if (ff > 0.0d+00) then
      x = xm
      go to 80
    end if
    dx = x

! do newton iteration until x converges to two decimal places
!
100 if (abs(dx / x) > 5.0d-03) then
      ff = pt(1)
      df = ff
      do i = 2, n
        ff = ff * x + pt(i)
        df = df * x + ff
      end do
      ff = ff * x + pt(nn)
      dx = ff / df
      x = x - dx
      go to 100
    end if
    bnd = x

! compute the derivative as the intial k polynomial and do 5 steps with no shift
!
    nm1 = n - 1
    do i = 2, n
      k(i) = (nn - i) * p(i) / n
    end do
    k(1) = p(1)
    aa   = p(nn)
    bb   = p(n)
    zerok = k(n) == 0.0d+00
    do jj = 1, 5
      cc = k(n)
      if (.not.zerok) then

! use scaled form of recurrence if value of k at 0 is nonzero
!
        t = - aa / cc
        do i = 1, nm1
          j    = nn - i
          k(j) = t * k(j - 1) + p(j)
        end do
        k(1) = p(1)
        zerok = abs(k(n)) <= abs(bb) * eta * 1.0d+01
      else

! use unscaled form of recurrence
!
        do i = 1, nm1
          j    = nn - i
          k(j) = k(j - 1)
        end do
        k(1) = 0.0d+00
        zerok = k(n) == 0.0d+00
      end if
    end do

! save k for restarts with new shifts
!
    temp(1:n) = k(1:n)

! loop to select the quadratic  corresponding to each new shift
!
    do cnt = 1, 20

! quadratic corresponds to a double shift to a non-real point and its complex
! conjugate.  the point has modulus bnd and amplitude rotated by 94 degrees
! from the previous shift
!
      xxx = cosr * xx - sinr * yy
      yy  = sinr * xx + cosr * yy
      xx  = xxx
      sr  = bnd * xx
      si  = bnd * yy
      u   = - 2.0d+00 * sr
      v   = bnd

! second stage calculation, fixed quadratic
!
      call fxshfr(20 * cnt,nz)

      if (nz /= 0) then

! the second stage jumps directly to one of the third stage iterations and
! returns here if successful; deflate the polynomial, store the zero or zeros
! and return to the main algorithm
!
        j          = degree - n + 1
        zeror(j)   = szr
        zeroi(j)   = szi
        nn         = nn - nz
        n          = nn - 1
        p(1:nn)    = qp(1:nn)
        if (nz == 1) go to 30
        zeror(j+1) = lzr
        zeroi(j+1) = lzi

        go to 30

      end if

! if the iteration is unsuccessful another quadratic is chosen after restoring k
!
      k(1:nn) = temp(1:nn)

    end do

! return with failure if no convergence with 20 shifts
!
    fail   = .true.
    degree = degree - n

!-------------------------------------------------------------------------------
!
  end function rpoly
!
!===============================================================================
!
! subroutine FXSHFR:
! -----------------
!
!   Computes up to  l2  fixed shift k-polynomials, testing for convergence in
!   the linear or quadratic case.  Initiates one of the variable shift
!   iterations and returns with the number of zeros found.
!
!   Arguments:
!
!      l2 - limit of fixed shift steps;
!      nz - number of zeros found;
!
!===============================================================================
!
  subroutine fxshfr(l2, nz)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer, intent(in)  :: l2
    integer, intent(out) :: nz

! local variables
!
    logical :: vpass, spass, vtry, stry
    integer :: tp, j, iflag
    real    :: svu, svv, ui, vi, s
    real    :: betas, betav, oss, ovv, ss, vv, ts, tv, ots, otv, tvv, tss
!
!-------------------------------------------------------------------------------
!
    nz    = 0
    betav = 0.25d+00
    betas = 0.25d+00
    oss   = sr
    ovv   = v

! evaluate polynomial by synthetic division
!
    call quadsd(nn, u, v, p, qp, a, b)
    call calcsc(tp)

    do j = 1, l2

! calculate next k polynomial and estimate v
!
      call nextk(tp)
      call calcsc(tp)
      call newest(tp, ui, vi)

      vv = vi

! estimate s
!
      ss = 0.0d+00
      if (k(n) /= 0.0d+00) ss = - p(nn) / k(n)
      tv = 1.0d+00
      ts = 1.0d+00
      if (j /= 1 .and. tp /= 3) then

! compute relative measures of convergence of s and v sequences
!
        if (vv /= 0.0d+00) tv = abs((vv - ovv) / vv)
        if (ss /= 0.0d+00) ts = abs((ss - oss) / ss)

! if decreasing, multiply two most recent convergence measures
!
        tvv = 1.0d+00
        if (tv < otv) tvv = tv * otv
        tss = 1.0d+00
        if (ts < ots) tss = ts * ots

! compare with convergence criteria
!
        vpass = tvv < betav
        spass = tss < betas

        if (spass .or. vpass) then

! at least one sequence has passed the convergence test.
! store variables before iterating
!
          svu = u
          svv = v
          svk(1:n) = k(1:n)
          s = ss

! choose iteration according to the fastest converging sequence
!
          vtry = .false.
          stry = .false.
          if (spass .and. ((.not.vpass) .or. tss < tvv)) go to 40
20        call quadit(ui, vi, nz)
          if (nz > 0) return

! quadratic iteration has failed. flag that it has
! been tried and decrease the convergence criterion.
!
          vtry = .true.
          betav = betav * 0.25d+00

! try linear iteration if it has not been tried and
! the s sequence is converging
!
          if (stry .or. (.not. spass)) go to 50
          k(1:n) = svk(1:n)
40        call realit(s, nz, iflag)
          if (nz > 0) return

! linear iteration has failed.  flag that it has been
! tried and decrease the convergence criterion
!
          stry = .true.
          betas = betas * 0.25d+00
          if (iflag /= 0) then

! if linear iteration signals an almost double real
! zero attempt quadratic interation
!
            ui = -(s + s)
            vi = s * s
            go to 20
          end if

! restore variables
!
50        u = svu
          v = svv
          k(1:n) = svk(1:n)

! try quadratic iteration if it has not been tried
! and the v sequence is converging
!
          if (vpass .and. (.not. vtry)) go to 20

! recompute qp and scalar values to continue the second stage
!
          call quadsd(nn, u, v, p, qp, a, b)
          call calcsc(tp)

        end if
      end if

      ovv = vv
      oss = ss
      otv = tv
      ots = ts

    end do

    return

!-------------------------------------------------------------------------------
!
  end subroutine fxshfr
!
!===============================================================================
!
! subroutine QUADIT:
! -----------------
!
!   Variable-shift k-polynomial iteration for a quadratic factor, converges
!   only if the zeros are equimodular or nearly so.
!
!   Arguments:
!
!      uu,vv - coefficients of starting quadratic;
!      nz    - number of zero found;
!
!===============================================================================
!
  subroutine quadit(uu, vv, nz)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    real   , intent(in)  :: uu
    real   , intent(in)  :: vv
    integer, intent(out) :: nz

! local variables
!
    logical :: tried
    integer :: tp, i, j
    real    :: ui, vi
    real    :: mp, omp, ee, relstp, t, zm
!
!-------------------------------------------------------------------------------
!
    nz = 0
    tried = .false.
    u = uu
    v = vv
    j = 0

! main loop
!
10  call quad(1.0d+00, u, v, szr, szi, lzr, lzi)

! return if roots of the quadratic are real and not close to multiple or nearly
! equal and of opposite sign.
!
    if (abs(abs(szr) - abs(lzr)) > 1.0d-02 * abs(lzr)) return

! evaluate polynomial by quadratic synthetic division
!
    call quadsd(nn, u, v, p, qp, a, b)
    mp = abs(a - szr * b) + abs(szi * b)

! compute a rigorous  bound on the rounding error in evaluting p
!
    zm = sqrt(abs(real(v)))
    ee = 2.0d+00 * abs(real(qp(1)))
    t  = - szr * b
    do i = 2, n
      ee = ee * zm + abs(real(qp(i)))
    end do
    ee = ee * zm + abs(real(a) + t)
    ee = (5.0d+00 * mre + 4.0d+00 * are) * ee - (5.0d+00 * mre+2.0d+00 * are)  &
       * (abs(real(a) + t) + abs(real(b)) * zm) + 2.0d+00 * are * abs(t)

! iteration has converged sufficiently if the polynomial value is less than
! 20 times this bound
!
    if (mp <= 2.0d+01 * ee) then
      nz = 2
      return
    end if
    j = j + 1

! stop iteration after 20 steps
!
    if (j > 20) return
    if (j >= 2) then
      if (.not.(relstp > 1.0d-02 .or. mp < omp .or. tried)) then

! a cluster appears to be stalling the convergence; five fixed shift steps are
! taken with a (u, v) close to the cluster
!
        if (relstp < eta) relstp = eta
        relstp = sqrt(relstp)
        u = u - u * relstp
        v = v + v * relstp
        call quadsd(nn, u, v, p, qp, a, b)
        do i = 1, 5
          call calcsc(tp)
          call nextk(tp)
        end do
        tried = .true.
        j = 0
      end if
    end if
    omp = mp

! calculate next k polynomial and new u and v
!
    call calcsc(tp)
    call nextk(tp)
    call calcsc(tp)
    call newest(tp, ui, vi)

! if vi is zero the iteration is not converging
!
    if (vi == 0.0d+00) return

    relstp = abs((vi - v) / vi)
    u = ui
    v = vi

    go to 10

!-------------------------------------------------------------------------------
!
  end subroutine quadit
!
!===============================================================================
!
! subroutine REALIT:
! -----------------
!
!   Variable-shift h polynomial iteration for a real zero.
!
!   Arguments:
!
!     sss   - starting iterate;
!     nz    - number of zero found;
!     iflag - flag to indicate a pair of zeros near real axis;
!
!===============================================================================
!
  subroutine realit(sss, nz, iflag)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    real   , intent(in out) :: sss
    integer, intent(out)    :: nz, iflag

! local variables
!
    integer :: i, j
    real    :: pv, kv, t, s
    real    :: ms, mp, omp, ee
!
!-------------------------------------------------------------------------------
!
    nz    = 0
    s     = sss
    iflag = 0
    j     = 0

! main loop
!
10  pv = p(1)

! evaluate p at s
!
    qp(1) = pv
    do i = 2, nn
      pv = pv * s + p(i)
      qp(i) = pv
    end do
    mp = abs(pv)

! compute a rigorous bound on the error in evaluating p
!
    ms = abs(s)
    ee = (mre / (are + mre)) * abs(real(qp(1)))
    do i = 2, nn
      ee = ee * ms + abs(real(qp(i)))
    end do

! iteration has converged sufficiently if the polynomial value is less than 20
! times this bound
!
    if (mp <= 2.0d+01 * ((are + mre) * ee - mre * mp)) then
      nz  = 1
      szr = s
      szi = 0.0d+00
      return
    end if
    j = j + 1

! stop iteration after 10 steps
!
    if (j > 10) return
    if (j >= 2) then
      if (abs(t) <= 1.0d-03 * abs(s - t) .and. mp > omp) then

! a cluster of zeros near the real axis has been encountered,
! return with iflag set to initiate a quadratic iteration
!
        iflag = 1
        sss   = s
        return
      end if
    end if

! return if the polynomial value has increased significantly
!
    omp = mp

! compute t, the next polynomial, and the new iterate
!
    kv    = k(1)
    qk(1) = kv
    do i = 2, n
      kv = kv * s + k(i)
      qk(i) = kv
    end do
    if (abs(kv) > abs(k(n)) * 1.0d+01 * eta) then

! use the scaled form of the recurrence if the value of k at s is nonzero
!
      t = -pv / kv
      k(1) = qp(1)
      do i = 2, n
        k(i) = t * qk(i - 1) + qp(i)
      end do
    else

! use unscaled form
!
      k(1) = 0.0d+00
      do i = 2, n
        k(i) = qk(i-1)
      end do
    end if

    kv = k(1)

    do i = 2, n
      kv = kv * s + k(i)
    end do

    t = 0.0d+00
    if (abs(kv) > abs(k(n)) * 1.0d+01 * eta) t = - pv / kv
    s = s + t

    go to 10

!-------------------------------------------------------------------------------
!
  end subroutine realit
!
!===============================================================================
!
! subroutine CALCSC:
! -----------------
!
!   Subroutine calculates scalar quantities used to compute the next k
!   polynomial and new estimates of the quadratic coefficients.
!
!   Arguments:
!
!     tp     - integer variable set here indicating how the calculations are
!              normalized to avoid overflow;
!
!===============================================================================
!
  subroutine calcsc(tp)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer, intent(out)  :: tp
!
!-------------------------------------------------------------------------------
!
! synthetic division of k by the quadratic 1,u,v
!
    call quadsd(n, u, v, k, qk, c, d)
    if (abs(c) <= abs(k(n)) * 1.0d+02 * eta) then
      if (abs(d) <= abs(k(n-1))*100.*eta) then
        tp = 3

! type=3 indicates the quadratic is almost a factor of k
!
        return

      end if
    end if

    if (abs(d) >= abs(c)) then
      tp = 2

! type=2 indicates that all formulas are divided by d
!
      e = a / d
      f = c / d
      g = u * b
      h = v * b
      a3 = (a + g) * e + h * (b / d)
      a1 = b * f - a
      a7 = (f + u) * a + h
      return

    end if

    tp = 1

! type=1 indicates that all formulas are divided by c
!
    e = a / c
    f = d / c
    g = u * e
    h = v * b
    a3 = a * e + (h/c+g) * b
    a1 = b - a * (d/c)
    a7 = a + g * d + h * f

    return

!-------------------------------------------------------------------------------
!
  end subroutine calcsc
!
!===============================================================================
!
! subroutine NEXTK:
! ----------------
!
!   Subroutine computes the next k polynomials using scalars computed in calcsc.
!
!   Arguments:
!
!     tp     - the type;
!
!===============================================================================
!
  subroutine nextk(tp)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer, intent(in)  :: tp

! local variables
!
    integer :: i
    real    :: temp
!
!-------------------------------------------------------------------------------
!
    if (tp /= 3) then

      temp = a
      if (tp == 1) temp = b
      if (abs(a1) <= abs(temp) * eta * 1.0d+01) then

! if a1 is nearly zero then use a special form of the recurrence
!
        k(1) = 0.0d+00
        k(2) = - a7 * qp(1)
        do i = 3, n
          k(i) = a3 * qk(i - 2) - a7 * qp(i - 1)
        end do

        return

      end if

! use scaled form of the recurrence
!
      a7   = a7 / a1
      a3   = a3 / a1
      k(1) = qp(1)
      k(2) = qp(2) - a7 * qp(1)
      do i = 3, n
        k(i) = a3 * qk(i - 2) - a7 * qp(i - 1) + qp(i)
      end do

      return

    end if

! use unscaled form of the recurrence if type is 3
!
    k(1) = 0.0d+00
    k(2) = 0.0d+00
    do i = 3, n
      k(i) = qk(i - 2)
    end do

    return

!-------------------------------------------------------------------------------
!
  end subroutine nextk
!
!===============================================================================
!
! subroutine NEWEST:
! -----------------
!
!   Subroutine computes new estimates of the quadratic coefficients using the
!   scalars computed in calcsc.
!
!   Arguments:
!
!     tp     - the type;
!     uu, vv - the estimates;
!
!===============================================================================
!
  subroutine newest(tp, uu, vv)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer, intent(in)  :: tp
    real   , intent(out) :: uu
    real   , intent(out) :: vv

! local variables
!
    real :: a4, a5, b1, b2, c1, c2, c3, c4, temp
!
!-------------------------------------------------------------------------------
!
! use formulas appropriate to setting of type.
!
    if (tp /= 3) then
      if (tp /= 2) then
        a4 = a + u * b + h * f
        a5 = c + (u + v * f) * d
      else
        a4 = (a + g) * f + h
        a5 = (f + u) * c + v * d
      end if

! evaluate new quadratic coefficients.
!
      b1 = - k(n) / p(nn)
      b2 = - (k(n-1) + b1 * p(n)) / p(nn)
      c1 = v * b2 * a1
      c2 = b1 * a7
      c3 = b1 * b1 * a3
      c4 = c1 - c2 - c3

      temp = a5 + b1 * a4 - c4
      if (temp /= 0.0d+00) then
        uu = u - (u * (c3 + c2) + v * (b1 * a1 + b2 * a7)) / temp
        vv = v * (1.0d+00 + c4 / temp)
        return
      end if
    end if

! if tp=3 the quadratic is zeroed
!
    uu = 0.0d+00
    vv = 0.0d+00

    return

!-------------------------------------------------------------------------------
!
  end subroutine newest
!
!===============================================================================
!
! subroutine QUADSD:
! -----------------
!
!   Subroutine divides p by the quadratic (1, u, v) placing the quotient in q
!   and the remainder in (a, b).
!
!   Arguments:
!
!     nn   - the degree of p;
!     u, v - the quadratic coefficients;
!     p    - the input polynomial coefficients;
!     q    - the divided polynomial coefficients;
!     a, b - the remainder of division;
!
!===============================================================================
!
  subroutine quadsd(nn, u, v, p, q, a, b)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer, intent(in)  :: nn
    real   , intent(in)  :: u, v, p(nn)
    real   , intent(out) :: q(nn), a, b

! local variables
!
    integer :: i
    real    :: c
!
!-------------------------------------------------------------------------------
!
    b    = p(1)
    q(1) = b
    a    = p(2) - u * b
    q(2) = a

    do i = 3, nn
      c    = p(i) - u * a - v * b
      q(i) = c
      b    = a
      a    = c
    end do

    return

!-------------------------------------------------------------------------------
!
  end subroutine quadsd
!
!===============================================================================
!
! subroutine QUAD:
! ---------------
!
!   Subroutine calculates the zeros of the quadratic equation:
!
!     f(x) = a x² + b₁ x + c = 0
!
!   Remark:
!
!     The quadratic formula, modified to avoid overflow, is used to find the
!     larger zero if the zeros are real and both zeros are complex.
!     The smaller real zero is found directly from the product of the zeros
!     c / a.
!
!   Arguments:
!
!     a, b₁, c - the quadratic equation coefficients;
!     sr, si   - the smaller zero's real and imaginary parts;
!     lr, li   - the larger zero's real and imaginary parts;
!
!===============================================================================
!
  subroutine quad(a, b1, c, sr, si, lr, li)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    real, intent(in)  :: a, b1, c
    real, intent(out) :: sr, si, lr, li

! local variables
!
    real :: b, d, e
!
!-------------------------------------------------------------------------------
!
    if (a /= 0.0d+00) go to 20
    sr = 0.0d+00
    if (b1 /= 0.0d+00) sr = - c / b1
    lr = 0.0d+00
10  si = 0.0d+00
    li = 0.0d+00
    return

20  if (c == 0.0d+00) then
      sr = 0.0d+00
      lr = - b1 / a
      go to 10
    end if

! compute discriminant avoiding overflow
!
    b = b1 / 2.0d+00
    if (abs(b) >= abs(c)) then
      e = 1.0d+00 - (a / b) * (c / b)
      d = sqrt(abs(e)) * abs(b)
    else
      e = a
      if (c < 0.0d+00) e = -a
      e = b * (b / abs(c)) - e
      d = sqrt(abs(e)) * sqrt(abs(c))
    end if

    if (e >= 0.0d+00) then

! real zeros
!
      if (b >= 0.0d+00) d = - d
      lr = ( - b + d ) / a
      sr = 0.0d+00
      if (lr /= 0.0d+00) sr = ( c / lr ) / a
      go to 10

    end if

! complex conjugate zeros
!
    sr = - b / a
    lr = sr
    si = abs(d / a)
    li = -si

    return

!-------------------------------------------------------------------------------
!
  end subroutine quad

!===============================================================================
!
end module algebra
